import * as T from "three";
const D = (
  /* glsl */
  `
    
#ifdef IS_VERTEX
    vec3 csm_Position;
    vec4 csm_PositionRaw;
    vec3 csm_Normal;

    // csm_PointSize
    #ifdef IS_POINTSMATERIAL
        float csm_PointSize;
    #endif
#else
    vec4 csm_DiffuseColor;
    vec4 csm_FragColor;
    float csm_FragColorInfluence;

    // csm_Emissive, csm_Roughness, csm_Metalness
    #if defined IS_MESHSTANDARDMATERIAL || defined IS_MESHPHYSICALMATERIAL
        vec3 csm_Emissive;
        float csm_Roughness;
        float csm_Metalness;
        float csm_Iridescence;
        
        #if defined IS_MESHPHYSICALMATERIAL
            float csm_Clearcoat;
            float csm_ClearcoatRoughness;
            vec3 csm_ClearcoatNormal;
            float csm_Transmission;
            float csm_Thickness;
        #endif
    #endif

    // csm_AO
    #if defined IS_MESHSTANDARDMATERIAL || defined IS_MESHPHYSICALMATERIAL || defined IS_MESHBASICMATERIAL || defined IS_MESHLAMBERTMATERIAL || defined IS_MESHPHONGMATERIAL || defined IS_MESHTOONMATERIAL
        float csm_AO;
    #endif

    // csm_Bump
    #if defined IS_MESHLAMBERTMATERIAL || defined IS_MESHMATCAPMATERIAL || defined IS_MESHNORMALMATERIAL || defined IS_MESHPHONGMATERIAL || defined IS_MESHPHYSICALMATERIAL || defined IS_MESHSTANDARDMATERIAL || defined IS_MESHTOONMATERIAL || defined IS_SHADOWMATERIAL 
        vec3 csm_Bump;
    #endif

    float csm_DepthAlpha;
#endif
`
), x = (
  /* glsl */
  `

#ifdef IS_VERTEX
    // csm_Position & csm_PositionRaw
    #ifdef IS_UNKNOWN
        csm_Position = vec3(0.0);
        csm_PositionRaw = vec4(0.0);
        csm_Normal = vec3(0.0);
    #else
        csm_Position = position;
        csm_PositionRaw = projectionMatrix * modelViewMatrix * vec4(position, 1.);
        csm_Normal = normal;
    #endif

    // csm_PointSize
    #ifdef IS_POINTSMATERIAL
        csm_PointSize = size;
    #endif
#else
    csm_FragColorInfluence = 0.0;

    // csm_DiffuseColor & csm_FragColor
    #if defined IS_UNKNOWN || defined IS_SHADERMATERIAL || defined IS_MESHDEPTHMATERIAL || defined IS_MESHNORMALMATERIAL || defined IS_SHADOWMATERIAL
        csm_DiffuseColor = vec4(1.0, 0.0, 1.0, 1.0);
        csm_FragColor = vec4(1.0, 0.0, 1.0, 1.0);
    #else
        #ifdef USE_MAP
            vec4 _csm_sampledDiffuseColor = texture2D(map, vMapUv);

            #ifdef DECODE_VIDEO_TEXTURE
            // inline sRGB decode (TODO: Remove this code when https://crbug.com/1256340 is solved)
            _csm_sampledDiffuseColor = vec4(mix(pow(_csm_sampledDiffuseColor.rgb * 0.9478672986 + vec3(0.0521327014), vec3(2.4)), _csm_sampledDiffuseColor.rgb * 0.0773993808, vec3(lessThanEqual(_csm_sampledDiffuseColor.rgb, vec3(0.04045)))), _csm_sampledDiffuseColor.w);
            #endif

            csm_DiffuseColor = vec4(diffuse, opacity) * _csm_sampledDiffuseColor;
            csm_FragColor = vec4(diffuse, opacity) * _csm_sampledDiffuseColor;
        #else
            csm_DiffuseColor = vec4(diffuse, opacity);
            csm_FragColor = vec4(diffuse, opacity);
        #endif
    #endif

    // csm_Emissive, csm_Roughness, csm_Metalness
    #if defined IS_MESHSTANDARDMATERIAL || defined IS_MESHPHYSICALMATERIAL
        csm_Emissive = emissive;
        csm_Roughness = roughness;
        csm_Metalness = metalness;

        #ifdef USE_IRIDESCENCE
            csm_Iridescence = iridescence;
        #else
            csm_Iridescence = 0.0;
        #endif

        #if defined IS_MESHPHYSICALMATERIAL
            #ifdef USE_CLEARCOAT
                csm_Clearcoat = clearcoat;
                csm_ClearcoatRoughness = clearcoatRoughness;
            #else
                csm_Clearcoat = 0.0;
                csm_ClearcoatRoughness = 0.0;
            #endif

            #ifdef USE_TRANSMISSION
                csm_Transmission = transmission;
                csm_Thickness = thickness;
            #else
                csm_Transmission = 0.0;
                csm_Thickness = 0.0;
            #endif
        #endif
    #endif

    // csm_AO
    #if defined IS_MESHSTANDARDMATERIAL || defined IS_MESHPHYSICALMATERIAL || defined IS_MESHBASICMATERIAL || defined IS_MESHLAMBERTMATERIAL || defined IS_MESHPHONGMATERIAL || defined IS_MESHTOONMATERIAL
        csm_AO = 0.0;
    #endif

    // csm_Bump
    #if defined IS_MESHLAMBERTMATERIAL || defined IS_MESHMATCAPMATERIAL || defined IS_MESHNORMALMATERIAL || defined IS_MESHPHONGMATERIAL || defined IS_MESHPHYSICALMATERIAL || defined IS_MESHSTANDARDMATERIAL || defined IS_MESHTOONMATERIAL || defined IS_SHADOWMATERIAL 
        csm_Bump = vec3(0.0);
    #endif

    csm_DepthAlpha = 1.0;
#endif
`
), y = (
  /* glsl */
  `
    varying mat4 csm_internal_vModelViewMatrix;
`
), b = (
  /* glsl */
  `
    csm_internal_vModelViewMatrix = modelViewMatrix;
`
), O = (
  /* glsl */
  `
    varying mat4 csm_internal_vModelViewMatrix;
`
), w = (
  /* glsl */
  `
    
`
), e = {
  // PBR
  diffuse: "csm_DiffuseColor",
  // Color + alpha
  normal: "csm_Normal",
  // Normal
  roughness: "csm_Roughness",
  // Roughness
  metalness: "csm_Metalness",
  // Metalness
  emissive: "csm_Emissive",
  // Emissive
  ao: "csm_AO",
  // AO
  bump: "csm_Bump",
  // Bump
  clearcoat: "csm_Clearcoat",
  // Clearcoat factor
  clearcoatRoughness: "csm_ClearcoatRoughness",
  // Clearcoat roughness
  clearcoatNormal: "csm_ClearcoatNormal",
  // Clearcoat normals
  transmission: "csm_Transmission",
  // Transmission
  thickness: "csm_Thickness",
  // Thickness
  iridescence: "csm_Iridescence",
  // Iridescence
  // Extras
  pointSize: "csm_PointSize",
  fragColor: "csm_FragColor",
  depthAlpha: "csm_DepthAlpha",
  // Depth
  fragColorInfluence: "csm_FragColorInfluence",
  // Unlit factor
  // Vert
  position: "csm_Position",
  positionRaw: "csm_PositionRaw"
}, k = {
  [`${e.position}`]: "*",
  [`${e.positionRaw}`]: "*",
  [`${e.normal}`]: "*",
  [`${e.pointSize}`]: ["PointsMaterial"],
  [`${e.diffuse}`]: "*",
  [`${e.fragColor}`]: "*",
  [`${e.fragColorInfluence}`]: ["*"],
  [`${e.emissive}`]: ["MeshStandardMaterial", "MeshPhysicalMaterial"],
  [`${e.roughness}`]: ["MeshStandardMaterial", "MeshPhysicalMaterial"],
  [`${e.metalness}`]: ["MeshStandardMaterial", "MeshPhysicalMaterial"],
  [`${e.iridescence}`]: [
    "MeshStandardMaterial",
    "MeshPhysicalMaterial"
  ],
  [`${e.ao}`]: [
    "MeshStandardMaterial",
    "MeshPhysicalMaterial",
    "MeshBasicMaterial",
    "MeshLambertMaterial",
    "MeshPhongMaterial",
    "MeshToonMaterial"
  ],
  [`${e.bump}`]: [
    "MeshLambertMaterial",
    "MeshMatcapMaterial",
    "MeshNormalMaterial",
    "MeshPhongMaterial",
    "MeshPhysicalMaterial",
    "MeshStandardMaterial",
    "MeshToonMaterial",
    "ShadowMaterial"
  ],
  [`${e.depthAlpha}`]: ["MeshDepthMaterial"],
  [`${e.clearcoat}`]: ["MeshPhysicalMaterial"],
  [`${e.clearcoatRoughness}`]: ["MeshPhysicalMaterial"],
  [`${e.clearcoatNormal}`]: ["MeshPhysicalMaterial"],
  [`${e.transmission}`]: ["MeshPhysicalMaterial"],
  [`${e.thickness}`]: ["MeshPhysicalMaterial"]
}, F = {
  // VERT
  "*": {
    "#include <lights_physical_fragment>": T.ShaderChunk.lights_physical_fragment,
    "#include <transmission_fragment>": T.ShaderChunk.transmission_fragment
  },
  [`${e.normal}`]: {
    "#include <beginnormal_vertex>": `
    vec3 objectNormal = ${e.normal};
    #ifdef USE_TANGENT
	    vec3 objectTangent = vec3( tangent.xyz );
    #endif
    `
  },
  [`${e.position}`]: {
    "#include <begin_vertex>": `
    vec3 transformed = ${e.position};
  `
  },
  [`${e.positionRaw}`]: {
    "#include <begin_vertex>": `
    vec4 csm_internal_positionUnprojected = ${e.positionRaw};
    mat4x4 csm_internal_unprojectMatrix = projectionMatrix * modelViewMatrix;
    #ifdef USE_INSTANCING
      csm_internal_unprojectMatrix = csm_internal_unprojectMatrix * instanceMatrix;
    #endif
    csm_internal_positionUnprojected = inverse(csm_internal_unprojectMatrix) * csm_internal_positionUnprojected;
    vec3 transformed = csm_internal_positionUnprojected.xyz;
  `
  },
  [`${e.pointSize}`]: {
    "gl_PointSize = size;": `
    gl_PointSize = ${e.pointSize};
    `
  },
  // FRAG
  [`${e.diffuse}`]: {
    "#include <color_fragment>": `
    #include <color_fragment>
    diffuseColor = ${e.diffuse};
  `
  },
  [`${e.fragColor}`]: {
    "#include <opaque_fragment>": `
    #include <opaque_fragment>
    gl_FragColor = mix(gl_FragColor, ${e.fragColor}, ${e.fragColorInfluence});
  `
  },
  [`${e.emissive}`]: {
    "vec3 totalEmissiveRadiance = emissive;": `
    vec3 totalEmissiveRadiance = ${e.emissive};
    `
  },
  [`${e.roughness}`]: {
    "#include <roughnessmap_fragment>": `
    #include <roughnessmap_fragment>
    roughnessFactor = ${e.roughness};
    `
  },
  [`${e.metalness}`]: {
    "#include <metalnessmap_fragment>": `
    #include <metalnessmap_fragment>
    metalnessFactor = ${e.metalness};
    `
  },
  [`${e.ao}`]: {
    "#include <aomap_fragment>": `
    #include <aomap_fragment>
    reflectedLight.indirectDiffuse *= 1. - ${e.ao};
    `
  },
  [`${e.bump}`]: {
    "#include <normal_fragment_maps>": `
    #include <normal_fragment_maps>

    vec3 csm_internal_orthogonal = ${e.bump} - (dot(${e.bump}, normal) * normal);
    vec3 csm_internal_projectedbump = mat3(csm_internal_vModelViewMatrix) * csm_internal_orthogonal;
    normal = normalize(normal - csm_internal_projectedbump);
    `
  },
  [`${e.depthAlpha}`]: {
    "gl_FragColor = vec4( vec3( 1.0 - fragCoordZ ), opacity );": `
      gl_FragColor = vec4( vec3( 1.0 - fragCoordZ ), opacity * ${e.depthAlpha} );
    `,
    "gl_FragColor = packDepthToRGBA( fragCoordZ );": `
      gl_FragColor = packDepthToRGBA( fragCoordZ );
      gl_FragColor.a *= ${e.depthAlpha};
    `
  },
  [`${e.clearcoat}`]: {
    "material.clearcoat = clearcoat;": `material.clearcoat = ${e.clearcoat};`
  },
  [`${e.clearcoatRoughness}`]: {
    "material.clearcoatRoughness = clearcoatRoughness;": `material.clearcoatRoughness = ${e.clearcoatRoughness};`
  },
  [`${e.clearcoatNormal}`]: {
    "#include <clearcoat_normal_fragment_begin>": `
      vec3 csm_coat_internal_orthogonal = csm_ClearcoatNormal - (dot(csm_ClearcoatNormal, nonPerturbedNormal) * nonPerturbedNormal);
      vec3 csm_coat_internal_projectedbump = mat3(csm_internal_vModelViewMatrix) * csm_coat_internal_orthogonal;
      vec3 clearcoatNormal = normalize(nonPerturbedNormal - csm_coat_internal_projectedbump);
    `
  },
  [`${e.transmission}`]: {
    "material.transmission = transmission;": `
      material.transmission = ${e.transmission};
    `
  },
  [`${e.thickness}`]: {
    "material.thickness = thickness;": `
      material.thickness = ${e.thickness};
    `
  },
  [`${e.iridescence}`]: {
    "material.iridescence = iridescence;": `
      material.iridescence = ${e.iridescence};
    `
  }
}, j = {
  clearcoat: [
    e.clearcoat,
    e.clearcoatNormal,
    e.clearcoatRoughness
  ],
  transmission: [e.transmission],
  iridescence: [e.iridescence]
};
function z(n) {
  let i = 0;
  for (let c = 0; c < n.length; c++)
    i = n.charCodeAt(c) + (i << 6) + (i << 16) - i;
  const o = i >>> 0;
  return String(o);
}
function B(n) {
  try {
    new n();
  } catch (i) {
    if (i.message.indexOf("is not a constructor") >= 0)
      return !1;
  }
  return !0;
}
function H(n, i) {
  for (const o in i)
    i[o] instanceof Object && Object.assign(i[o], H(n[o], i[o]));
  return Object.assign(n || {}, i), n;
}
function P(n) {
  return n.replace(/\/\*[\s\S]*?\*\/|\/\/.*/g, "");
}
function U(n) {
  return n.replace(/\n/g, "");
}
function V(n) {
  return n.replace(/\s/g, "");
}
function G(n) {
  const i = n.name;
  return V(U(n.toString())) === `${i}(){}`;
}
class Y extends T.Material {
  constructor({
    baseMaterial: i,
    vertexShader: o,
    fragmentShader: M,
    uniforms: c,
    patchMap: R,
    cacheKey: _,
    ...l
  }) {
    if (!i)
      throw new Error("CustomShaderMaterial: baseMaterial is required.");
    let s;
    if (B(i)) {
      const E = Object.keys(l).length === 0;
      s = new i(E ? void 0 : l);
    } else
      s = i, Object.assign(s, l);
    if (["ShaderMaterial", "RawShaderMaterial"].includes(s.type))
      throw new Error(
        `CustomShaderMaterial does not support ${s.type} as a base material.`
      );
    super(), this.uniforms = {}, this.vertexShader = "", this.fragmentShader = "";
    const m = s;
    return m.name = `CustomShaderMaterial<${s.name}>`, m.update = this.update.bind(m), m.uniforms = this.uniforms = c || {}, m.vertexShader = this.vertexShader = o || "", m.fragmentShader = this.fragmentShader = M || "", m.update({
      fragmentShader: M,
      vertexShader: o,
      uniforms: c,
      patchMap: R,
      cacheKey: _
    }), m;
  }
  update({
    fragmentShader: i,
    vertexShader: o,
    uniforms: M,
    cacheKey: c,
    patchMap: R
  }) {
    const _ = P(o || ""), l = P(i || ""), s = this;
    M && (s.uniforms = M), o && (s.vertexShader = o), i && (s.fragmentShader = i), Object.entries(j).forEach(([a, d]) => {
      for (const h in d) {
        const t = d[h];
        (l && l.includes(t) || _ && _.includes(t)) && (s[a] || (s[a] = 1));
      }
    });
    const S = s.onBeforeCompile, m = !G(S), E = (a, d, h) => {
      var A;
      let t, f = "";
      if (d) {
        const r = /void\s+main\s*\(\s*\)[^{]*{((?:[^{}]+|{(?:[^{}]+|{(?:[^{}]+|{(?:[^{}]+|{[^{}]*})*})*})*})*})/gm;
        t = (A = d.matchAll(r).next().value) == null ? void 0 : A[1], t && (t = t.slice(0, -1));
        const v = d.indexOf("void main() {");
        f = d.slice(0, v);
      }
      if (m) {
        a = a.replace(
          "void main() {",
          `
          // THREE-CustomShaderMaterial by Faraz Shaikh: https://github.com/FarazzShaikh/THREE-CustomShaderMaterial
  
          ${f}
          
          void main() {
          `
        );
        const r = a.lastIndexOf("//~CSM_MAIN_END");
        if (r !== -1) {
          const I = `
            ${t ? `${t}` : ""}
            //~CSM_MAIN_END
          `;
          a = a.slice(0, r) + I + a.slice(r);
        }
      } else
        a = a.replace(
          "void main() {",
          `
          // THREE-CustomShaderMaterial by Faraz Shaikh: https://github.com/FarazzShaikh/THREE-CustomShaderMaterial
  
          ${h ? O : y}
          ${D}
  
          ${f}
          
          void main() {
            ${x}
            ${h ? w : b}

            ${t ? `${t}` : ""}
            //~CSM_MAIN_END
          `
        );
      return a;
    };
    s.onBeforeCompile = (a, d) => {
      S == null || S(a, d);
      const t = H(F, R || {}), f = s.type, A = f ? `#define IS_${f.toUpperCase()};
` : `#define IS_UNKNOWN;
`;
      a.vertexShader = A + `#define IS_VERTEX
` + a.vertexShader, a.fragmentShader = A + `#define IS_FRAGMENT
` + a.fragmentShader;
      for (const r in t) {
        const I = r === "*" || _ && _.includes(r);
        if (r === "*" || l && l.includes(r) || I) {
          const p = k[r];
          if (p && p !== "*" && (Array.isArray(p) ? !p.includes(f) : p !== f)) {
            console.error(
              `CustomShaderMaterial: ${r} is not available in ${f}. Shader cannot compile.`
            );
            return;
          }
          const $ = t[r];
          for (const g in $) {
            const u = $[g];
            if (typeof u == "object") {
              const N = u.type, L = u.value;
              N === "fs" ? a.fragmentShader = a.fragmentShader.replace(
                g,
                L
              ) : N === "vs" && (a.vertexShader = a.vertexShader.replace(
                g,
                L
              ));
            } else
              u && (a.vertexShader = a.vertexShader.replace(
                g,
                u
              ), a.fragmentShader = a.fragmentShader.replace(
                g,
                u
              ));
          }
        }
      }
      a.vertexShader = E(
        a.vertexShader,
        _,
        !1
      ), a.fragmentShader = E(
        a.fragmentShader,
        l,
        !0
      ), M && (a.uniforms = { ...a.uniforms, ...s.uniforms }), s.uniforms = a.uniforms;
    };
    const C = s.customProgramCacheKey;
    s.customProgramCacheKey = () => ((c == null ? void 0 : c()) || z((_ || "") + (l || ""))) + (C == null ? void 0 : C.call(s)), s.needsUpdate = !0;
  }
}
export {
  Y as default
};
//# sourceMappingURL=vanilla.js.map
